////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// DirectDrawPanel.pas - DirectDraw panel with back buffer
// -------------------------------------------------------
// Changed:   2005-01-09
// Maintain:  Michael Vinther: http://logicnet.dk/lib   |   mv@logicnet.dk
//
// Last change:
//   WaitForVerticalBlank in Draw
//
unit DirectDrawPanel;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DirectDrawUtils, DirectX, FastBitmap, Math, LinarBitmap;

type
  TDirectDrawPanel = class(TWinControl)
  private
    { Private declarations }
    PaintFX : DDBLTFX;
    FClientRect : TRect;
    FWaitForVerticalBlank : Boolean;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
    procedure WMEraseBkgnd(var Msg: TWMEraseBkgnd); message WM_ERASEBKGND;
  protected
    { Protected declarations }
    Stretch : Boolean;
    BufferWidth, BufferHeight : Integer;
    FBackBuffer : IDirectDrawSurface7;
    procedure UpdateDirectDraw;
    procedure Resize; override;
  public
    { Public declarations }
    property DDSBackBuffer: IDirectDrawSurface7 read FBackBuffer;
    property ClientRect: TRect read FClientRect;
    constructor Create(AOwner: TComponent); override;
    procedure Paint; virtual;
    procedure PaintBitmap(Bitmap: TFastBitmap);
    function GetDirectDrawSurface: IDirectDrawSurface7;
    procedure StretchBackBuffer(SourceWidth: Integer=-1; SourceHeight: Integer=-1); // -1 for no stretch
  published
    { Published declarations }
    property WaitForVerticalBlank: Boolean read FWaitForVerticalBlank write FWaitForVerticalBlank default True;
    property Align;
    property Anchors;
    property AutoSize;
    {property BevelInner;
    property BevelOuter;
    property BevelWidth;
    property BorderWidth;}
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    //property Font;
    property ParentBiDiMode;
    //property ParentColor;
    //property ParentCtl3D;
    //property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    //property TabOrder;
    //property TabStop;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    //property OnDockDrop;
    //property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    //property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    //property OnUnDock;
  end;

  TDirectDrawControl = class(TControl)
  private
    { Private declarations }
    PaintFX : DDBLTFX;
    FClientRect : TRect;
    FWaitForVerticalBlank : Boolean;
    procedure WMPaint(var Message: TWMPaint); message WM_PAINT;
  protected
    { Protected declarations }
    Stretch : Boolean;
    FBufferWidth, FBufferHeight : Integer;
    FBackBuffer : IDirectDrawSurface7;
    procedure UpdateDirectDraw;
    procedure Resize; override;
  public
    { Public declarations }
    property DDSBackBuffer: IDirectDrawSurface7 read FBackBuffer;
    property ClientRect: TRect read FClientRect;
    property BufferWidth: Integer read FBufferWidth;
    property BufferHeight: Integer read FBufferHeight;
    constructor Create(AOwner: TComponent); override;
    procedure Paint; virtual;
    procedure StretchBackBuffer(SourceWidth: Integer=-1; SourceHeight: Integer=-1); // -1 for no stretch
  published
    { Published declarations }
    property WaitForVerticalBlank: Boolean read FWaitForVerticalBlank write FWaitForVerticalBlank default True;
    property Align;
    property Anchors;
    property AutoSize;
    property Constraints;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;
    property ParentBiDiMode;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnCanResize;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
  end;

var
  DirectDraw : IDirectDraw7;
  DDSPrimary : IDirectDrawSurface7;

procedure Register;
function CreateDirectDrawSurface(Width,Height: Integer; PixelFormat: TPixelFormat=pfCustom): IDirectDrawSurface7;

implementation

{$R DirectDrawPanel.res}

uses MemUtils;

procedure Register;
begin
  RegisterComponents('Samples', [TDirectDrawPanel]);
  RegisterComponents('Samples', [TDirectDrawControl]);
end;

function CreateDirectDrawSurface(Width,Height: Integer; PixelFormat: TPixelFormat): IDirectDrawSurface7;
var
  SurfaceDesc : TDDSurfaceDesc2;  // Surface description
begin
  ZeroMem(SurfaceDesc,SizeOf(SurfaceDesc));
  with SurfaceDesc do
  begin
    dwSize:=SizeOf(SurfaceDesc);
    dwFlags:=DDSD_CAPS or DDSD_WIDTH or DDSD_HEIGHT;  // Valid fields in SurfaceDesc
    ddsCaps.dwCaps:=DDSCAPS_OFFSCREENPLAIN;
    dwWidth:=Width;
    dwHeight:=Height;
    if PixelFormat<>pfCustom then
    begin
      dwFlags:=dwFlags or DDSD_PIXELFORMAT;
      with ddpfPixelFormat do
      begin
        dwSize:=SizeOf(TDDPixelFormat);
        dwFlags:=DDPF_RGB;
        dwRBitMask:=$0000ff;
        dwGBitMask:=$00ff00;
        dwBBitMask:=$ff0000;
        case PixelFormat of
          //pf8bit  : dwRGBBitCount:=8;
          pf24bit : dwRGBBitCount:=24;
          pf32bit : dwRGBBitCount:=32;
        else
          raise Exception.Create(rsInvalidPixelFormat);
        end;
      end;
    end;
  end;
  DDFailed(DirectDraw.CreateSurface(SurfaceDesc,Result,nil));
end;

procedure InitDirectDraw;
var
  SurfaceDesc : TDDSurfaceDesc2;
begin
  if DirectDraw=nil then // Create DirectDraw interface
  begin
    DirectDraw:=GetDirectDraw7Interface;
    DDFailed(DirectDraw.SetCooperativeLevel(Application.Handle,DDSCL_NORMAL));
  end;
  if DDSPrimary=nil then // Create the primary surface
  begin
    ZeroMem(SurfaceDesc,SizeOf(SurfaceDesc));
    with SurfaceDesc do
    begin
      dwSize:=SizeOf(SurfaceDesc);
      dwFlags:=DDSD_CAPS;  // Valid fields in SurfaceDesc
      ddsCaps.dwCaps:=DDSCAPS_PRIMARYSURFACE;
    end;
    DDFailed(DirectDraw.CreateSurface(SurfaceDesc,DDSPrimary,nil));
  end;
end;

type
  TOpenWinControl = class(TWinControl);

//==============================================================================================================================
// TDirectDrawPanel
//==============================================================================================================================

constructor TDirectDrawPanel.Create(AOwner: TComponent);
begin
  Width:=100; Height:=100;
  BufferWidth:=Width; BufferHeight:=Height;
  PaintFX.dwSize:=SizeOf(PaintFX);
  FWaitForVerticalBlank:=True;
  inherited;
  if not (csDesigning in ComponentState) then InitDirectDraw;
end;

procedure TDirectDrawPanel.UpdateDirectDraw;
var
  SurfaceDesc : TDDSurfaceDesc2;  // Surface description
  Clipper : IDirectDrawClipper;
begin
  // Create back buffer
  if not Stretch or (FBackBuffer=nil) then
  begin
    FBackBuffer:=nil;
    ZeroMem(SurfaceDesc,SizeOf(SurfaceDesc));
    with SurfaceDesc do
    begin
      dwSize:=SizeOf(SurfaceDesc);
      dwFlags:=DDSD_CAPS or DDSD_WIDTH or DDSD_HEIGHT;  // Valid fields in SurfaceDesc
      ddsCaps.dwCaps:=DDSCAPS_OFFSCREENPLAIN;
      dwWidth:=Max(1,BufferWidth);
      dwHeight:=Max(1,BufferHeight);
    end;
    DDFailed(DirectDraw.CreateSurface(SurfaceDesc,FBackBuffer,nil));
  end;
  FClientRect.Right:=Width;
  FClientRect.Bottom:=Height;

  // Set clipper
  try
    DDFailed(DirectDraw.CreateClipper(0,Clipper,nil));
    DDFailed(Clipper.SetHWnd(0,WindowHandle));
    DDSPrimary.SetClipper(Clipper);
  finally
    Clipper:=nil;
  end;
end;

procedure TDirectDrawPanel.Resize;
begin
  if not (csDesigning in ComponentState) then
  begin
    if not Stretch then
    begin
      BufferWidth:=Width;
      BufferHeight:=Height;
    end;
    UpdateDirectDraw;
  end;
  if Assigned(OnResize) then OnResize(Self);
end;

procedure TDirectDrawPanel.WMPaint(var Message: TWMPaint);
begin
  inherited;
  Paint;
end;

procedure TDirectDrawPanel.WMEraseBkgnd(var Msg : TWMEraseBkgnd);
begin
  if csDesigning in ComponentState then inherited
  else Msg.Result:=LRESULT(False);
end;

procedure TDirectDrawPanel.Paint;
begin
  if Assigned(DDSBackBuffer) then
  begin
    if FWaitForVerticalBlank then
      DirectDraw.WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,0);
    DDSPrimary.Blt(Bounds(ClientOrigin.X,ClientOrigin.Y,Width,Height),DDSBackBuffer,TRect(nil^),DDBLT_WAIT,PaintFX);
  end;
end;

procedure TDirectDrawPanel.PaintBitmap(Bitmap: TFastBitmap);
var
  SurfaceDesc : TDDSurfaceDesc2;
  Y, BitmapMaxY, BytesPerLine : Integer;
begin
  Assert(Bitmap.PixelFormat=pf32bit);

  SurfaceDesc.dwSize:=SizeOf(SurfaceDesc);
  DDFailed(DDSBackBuffer.Lock(nil,SurfaceDesc,DDLOCK_WAIT or DDLOCK_WRITEONLY,0));

  BytesPerLine:=Min(Bitmap.BytesPerLine,SurfaceDesc.dwWidth*4);
  BitmapMaxY:=Bitmap.Height-1;
  for Y:=0 to Min(SurfaceDesc.dwHeight-1,BitmapMaxY) do
    Move(Bitmap.ScanLine32[BitmapMaxY-Y]^,Pointer(Integer(SurfaceDesc.lpSurface)+Y*SurfaceDesc.lPitch)^,BytesPerLine);
  DDFailed(DDSBackBuffer.UnLock(nil));

  Paint;
end;

function TDirectDrawPanel.GetDirectDrawSurface: IDirectDrawSurface7;
var
  SurfaceDesc : TDDSurfaceDesc2;  // Surface description
begin
  SurfaceDesc.dwSize:=SizeOf(SurfaceDesc);
  with SurfaceDesc do
  begin
    dwFlags:=DDSD_CAPS or DDSD_WIDTH or DDSD_HEIGHT;  // Valid fields in SurfaceDesc
    ddsCaps.dwCaps:=DDSCAPS_OFFSCREENPLAIN;
    dwWidth:=Max(1,BufferWidth);
    dwHeight:=Max(1,BufferHeight);
  end;
  DDFailed(DirectDraw.CreateSurface(SurfaceDesc,Result,nil));
end;

procedure TDirectDrawPanel.StretchBackBuffer(SourceWidth,SourceHeight: Integer); // -1 for no stretch
begin
  if Width=-1 then
  begin
    Stretch:=False;
    BufferWidth:=Width;
    BufferHeight:=Height;
    UpdateDirectDraw;
  end
  else
  begin
    BufferWidth:=SourceWidth;
    BufferHeight:=SourceHeight;
    Stretch:=False;
    UpdateDirectDraw;
    Stretch:=True;
  end;
end;

//==============================================================================================================================
// TDirectDrawControl
//==============================================================================================================================

constructor TDirectDrawControl.Create(AOwner: TComponent);
begin
  FBufferWidth:=Width; FBufferHeight:=Height;
  PaintFX.dwSize:=SizeOf(PaintFX);
  FWaitForVerticalBlank:=True;
  if not (csDesigning in ComponentState) then InitDirectDraw;
  inherited;
end;

procedure TDirectDrawControl.UpdateDirectDraw;
var
  SurfaceDesc : TDDSurfaceDesc2;  // Surface description
  Clipper : IDirectDrawClipper;
begin
  // Create back buffer
  if not Stretch or (FBackBuffer=nil) then
  begin
    FBackBuffer:=nil;
    ZeroMem(SurfaceDesc,SizeOf(SurfaceDesc));
    with SurfaceDesc do
    begin
      dwSize:=SizeOf(SurfaceDesc);
      dwFlags:=DDSD_CAPS or DDSD_WIDTH or DDSD_HEIGHT;  // Valid fields in SurfaceDesc
      ddsCaps.dwCaps:=DDSCAPS_OFFSCREENPLAIN;
      dwWidth:=Max(1,BufferWidth);
      dwHeight:=Max(1,BufferHeight);
    end;
    DDFailed(DirectDraw.CreateSurface(SurfaceDesc,FBackBuffer,nil));
  end;
  FClientRect.Right:=Width;
  FClientRect.Bottom:=Height;

  // Set clipper
  try
    DDFailed(DirectDraw.CreateClipper(0,Clipper,nil));
    DDFailed(Clipper.SetHWnd(0,TOpenWinControl(Parent).WindowHandle));
    DDSPrimary.SetClipper(Clipper);
  finally
    Clipper:=nil;
  end;
end;

procedure TDirectDrawControl.Resize;
begin
  if not (csDesigning in ComponentState) then
  begin
    if not Stretch then
    begin
      FBufferWidth:=Width;
      FBufferHeight:=Height;
    end;
    UpdateDirectDraw;
  end;
  if Assigned(OnResize) then OnResize(Self);
end;

procedure TDirectDrawControl.WMPaint(var Message: TWMPaint);
begin
  inherited;
  Paint;
end;

procedure TDirectDrawControl.Paint;
begin
  if Assigned(DDSBackBuffer) then
  begin
    if FWaitForVerticalBlank then
      DirectDraw.WaitForVerticalBlank(DDWAITVB_BLOCKBEGIN,0);
    DDSPrimary.Blt(Bounds(ClientOrigin.X,ClientOrigin.Y,Width,Height),DDSBackBuffer,TRect(nil^),DDBLT_WAIT,PaintFX);
  end;
end;

procedure TDirectDrawControl.StretchBackBuffer(SourceWidth,SourceHeight: Integer); // -1 for no stretch
begin
  if Width=-1 then
  begin
    Stretch:=False;
    FBufferWidth:=Width;
    FBufferHeight:=Height;
    UpdateDirectDraw;
  end
  else
  begin
    FBufferWidth:=SourceWidth;
    FBufferHeight:=SourceHeight;
    Stretch:=False;
    UpdateDirectDraw;
    Stretch:=True;
  end;
end;

end.

