////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// DirectDrawUtils.pas - DirectDraw utilities
// ------------------------------------------
// Changed:   2002-09-24
// Maintain:  Michael Vinther  |  mv@logicnet.dk
//
// Last change:
//
unit DirectDrawUtils;

interface

uses Windows, SysUtils, ActiveX, DirectX, MemUtils, LinarBitmap, Graphics;

type EDirectDrawError = class(Exception);

function GetDirectDrawInterface: IDirectDraw;
function GetDirectDraw4Interface: IDirectDraw4;
function GetDirectDraw7Interface: IDirectDraw7;

procedure RenderFrameGray(Frame: PByteArray; FrameBuffer: IDirectDrawSurface7);
procedure RenderFrame24Flip(Frame: PByteArray; FrameBuffer: IDirectDrawSurface7);
procedure DDSurfaceFillColor(const Surface: IDirectDrawSurface7; Color: DWord);

// Create bitmap from contents of surface
procedure DrawSurfaceToLinearBitmapGray(Surface: IDirectDrawSurface7; Image: TLinearBitmap);
procedure DrawSurfaceToLinearBitmapRGB(Surface: IDirectDrawSurface7; Image: TLinearBitmap);

procedure CreateInterfaceFailed(Result: DWord);
procedure DDFailed(Result: Integer);
function DDErrorString(Error: Integer): string;

implementation

procedure RenderFrameGray(Frame: PByteArray; FrameBuffer: IDirectDrawSurface7);
var
  X, Y : Integer;
  SrcPix : PByte;
  DestPix : PDWord;
  DestPix24 : PRGBRec;
  SurfaceDesc : TDDSurfaceDesc2;  // Surface description
begin
  SrcPix:=Pointer(Frame);
  SurfaceDesc.dwSize:=SizeOf(SurfaceDesc);
  if FrameBuffer.Lock(nil,SurfaceDesc,DDLOCK_WAIT or DDLOCK_WRITEONLY,0)=DD_OK then
  begin
    if SurfaceDesc.ddpfPixelFormat.dwRGBBitCount=32 then
    begin
      for Y:=0 to SurfaceDesc.dwHeight-1 do
      begin
        DestPix:=Pointer(Integer(SurfaceDesc.lpSurface)+Y*SurfaceDesc.lPitch);
        for X:=1 to SurfaceDesc.dwWidth do
        begin
          //PByte(DWord(DestPix)+0)^:=SrcPix^; PByte(DWord(DestPix)+1)^:=SrcPix^; PByte(DWord(DestPix)+2)^:=SrcPix^;
          DestPix^:=SrcPix^ or (SrcPix^ shl 8) or (SrcPix^ shl 16);
          Inc(SrcPix);
          Inc(DestPix);
        end;
      end
    end
    else if SurfaceDesc.ddpfPixelFormat.dwRGBBitCount=24 then
    begin
      for Y:=0 to SurfaceDesc.dwHeight-1 do
      begin
        DestPix24:=Pointer(Integer(SurfaceDesc.lpSurface)+Y*SurfaceDesc.lPitch);
        for X:=1 to SurfaceDesc.dwWidth do
        begin
          DestPix24^.B:=SrcPix^; DestPix24^.G:=SrcPix^; DestPix24^.R:=SrcPix^;
          Inc(SrcPix);
          Inc(DestPix24);
        end;
      end;
    end;
    FrameBuffer.UnLock(nil);
  end;
end;

procedure RenderFrame24Flip(Frame: PByteArray; FrameBuffer: IDirectDrawSurface7);
var
  X, Y : Integer;
  SrcPix : PRGBRec;
  DestPix : PDWord;
  SurfaceDesc : TDDSurfaceDesc2;  // Surface description
begin
  SrcPix:=Pointer(Frame);
  SurfaceDesc.dwSize:=SizeOf(SurfaceDesc);
  if FrameBuffer.Lock(nil,SurfaceDesc,DDLOCK_WAIT or DDLOCK_WRITEONLY,0)=DD_OK then
  begin
    if SurfaceDesc.ddpfPixelFormat.dwRGBBitCount=32 then
      for Y:=SurfaceDesc.dwHeight-1 downto 0 do
      begin
        DestPix:=Pointer(Integer(SurfaceDesc.lpSurface)+Y*SurfaceDesc.lPitch);
        for X:=1 to SurfaceDesc.dwWidth do
        begin
          DestPix^:=PDWord(SrcPix)^;
          Inc(SrcPix);
          Inc(DestPix);
        end;
      end
    else if SurfaceDesc.ddpfPixelFormat.dwRGBBitCount=24 then
    begin
      X:=SurfaceDesc.dwWidth*3;
      for Y:=SurfaceDesc.dwHeight-1 downto 0 do
      begin
        Move(SrcPix^,Pointer(Integer(SurfaceDesc.lpSurface)+Y*SurfaceDesc.lPitch)^,X);
        Inc(PByte(SrcPix),X);
      end
    end;
    FrameBuffer.UnLock(nil);
  end;
end;

function DDErrorString(Error: Integer): string;
begin
  case Error of
    DD_OK : Result:='The request completed successfully';
    DDERR_ALREADYINITIALIZED : Result:='The object has already been initialized';
    DDERR_BLTFASTCANTCLIP : Result:='A DirectDrawClipper object is attached to a source surface that has passed into a call to the IDirectDrawSurface7::BltFast method';
    DDERR_CANNOTATTACHSURFACE : Result:='A surface cannot be attached to another requested surface';
    DDERR_CANNOTDETACHSURFACE : Result:='A surface cannot be detached from another requested surface';
    DDERR_CANTCREATEDC : Result:='Windows cannot create any more device contexts (DCs), or a DC has requested a palette-indexed surface when the surface had no palette and the display mode was not palette-indexed (in this case DirectDraw cannot select a proper palette into the DC)';
    DDERR_CANTDUPLICATE : Result:='Primary and 3-D surfaces, or surfaces that are implicitly created, cannot be duplicated';
    DDERR_CANTLOCKSURFACE : Result:='Access to this surface is refused because an attempt was made to lock the primary surface without DCI support';
    DDERR_CANTPAGELOCK : Result:='An attempt to page-lock a surface failedPage lock does not work on a display-memory surface or an emulated primary surface';
    DDERR_CANTPAGEUNLOCK : Result:='An attempt to page-unlock a surface failedPage unlock does not work on a display-memory surface or an emulated primary surface';
    DDERR_CLIPPERISUSINGHWND : Result:='An attempt was made to set a clip list for a DirectDrawClipper object that is already monitoring a window handle';
    DDERR_COLORKEYNOTSET : Result:='No source color key is specified for this operation';
    DDERR_CURRENTLYNOTAVAIL : Result:='No support is currently available';
    DDERR_DDSCAPSCOMPLEXREQUIRED : Result:='The surface requires the DDSCAPS_COMPLEX flag';
    DDERR_DCALREADYCREATED : Result:='A device context (DC) has already been returned for this surfaceOnly one DC can be retrieved for each surface';
    DDERR_DEVICEDOESNTOWNSURFACE : Result:='Surfaces created by one DirectDraw device cannot be used directly by another DirectDraw device';
    DDERR_DIRECTDRAWALREADYCREATED : Result:='A DirectDraw object representing this driver has already been created for this process';
    DDERR_EXCEPTION : Result:='An exception was encountered while performing the requested operation';
    DDERR_EXCLUSIVEMODEALREADYSET : Result:='An attempt was made to set the cooperative level when it was already set to exclusive';
    DDERR_EXPIRED : Result:='The data has expired and is therefore no longer valid';
    DDERR_GENERIC : Result:='There is an undefined error condition';
    DDERR_HEIGHTALIGN : Result:='The height of the provided rectangle is not a multiple of the required alignment';
    DDERR_HWNDALREADYSET : Result:='The DirectDraw cooperative-level window handle has already been setIt cannot be reset while the process has surfaces or palettes created';
    DDERR_HWNDSUBCLASSED : Result:='DirectDraw is prevented from restoring state because the DirectDraw cooperative-level window handle has been subclassed';
    DDERR_IMPLICITLYCREATED : Result:='The surface cannot be restored because it is an implicitly created surface';
    DDERR_INCOMPATIBLEPRIMARY : Result:='The primary surface creation request does not match the existing primary surface';
    DDERR_INVALIDCAPS : Result:='One or more of the capability bits passed to the callback function are incorrect';
    DDERR_INVALIDCLIPLIST : Result:='DirectDraw does not support the provided clip list';
    DDERR_INVALIDDIRECTDRAWGUID : Result:='The globally unique identifier (GUID) passed to the DirectDrawCreate function is not a valid DirectDraw driver identifier';
    DDERR_INVALIDMODE : Result:='DirectDraw does not support the requested mode';
    DDERR_INVALIDOBJECT : Result:='DirectDraw received a pointer that was an invalid DirectDraw object';
    DDERR_INVALIDPARAMS : Result:='One or more of the parameters passed to the method are incorrect';
    DDERR_INVALIDPIXELFORMAT : Result:='The pixel format was invalid as specified';
    DDERR_INVALIDPOSITION : Result:='The position of the overlay on the destination is no longer legal';
    DDERR_INVALIDRECT : Result:='The provided rectangle was invalid';
    DDERR_INVALIDSTREAM : Result:='The specified stream contains invalid data';
    DDERR_INVALIDSURFACETYPE : Result:='The surface was of the wrong type';
    DDERR_LOCKEDSURFACES : Result:='One or more surfaces are locked, causing the failure of the requested operation';
    DDERR_MOREDATA : Result:='There is more data available than the specified buffer size can hold';
    DDERR_NEWMODE : Result:='When IDirectDraw7::StartModeTest is called with the DDSMT_ISTESTREQUIRED flag, it may return this value to denote that some or all of the resolutions can and should be tested.'+
                            'IDirectDraw7::EvaluateMode returns this value to indicate that the test has switched to a new display mode';
    DDERR_NO3D : Result:='No 3-D hardware or emulation is present';
    DDERR_NOALPHAHW : Result:='No alpha-acceleration hardware is present or available, causing the failure of the requested operation';
    DDERR_NOBLTHW : Result:='No blitter hardware is present';
    DDERR_NOCLIPLIST : Result:='No clip list is available';
    DDERR_NOCLIPPERATTACHED : Result:='No DirectDrawClipper object is attached to the surface object';
{DDERR_NOCOLORCONVHW
No color-conversion hardware is present or available.
DDERR_NOCOLORKEY
The surface does not currently have a color key.
DDERR_NOCOLORKEYHW
There is no hardware support for the destination color key.
DDERR_NOCOOPERATIVELEVELSET
A create function was called without the IDirectDraw7::SetCooperativeLevel method.
DDERR_NODC
No device context (DC) has ever been created for this surface.
DDERR_NODDROPSHW
No DirectDraw raster-operation (ROP) hardware is available. 
DDERR_NODIRECTDRAWHW
Hardware-only DirectDraw object creation is not possible; the driver does not support any hardware. 
DDERR_NODIRECTDRAWSUPPORT
DirectDraw support is not possible with the current display driver.
DDERR_NODRIVERSUPPORT 
New for DirectX 7.0. Testing cannot proceed because the display adapter driver does not enumerate refresh rates. 
DDERR_NOEMULATION 
Software emulation is not available. 
DDERR_NOEXCLUSIVEMODE 
The operation requires the application to have exclusive mode, but the application does not have exclusive mode. 
DDERR_NOFLIPHW 
Flipping visible surfaces is not supported. 
DDERR_NOFOCUSWINDOW
An attempt was made to create or set a device window without first setting the focus window.
DDERR_NOGDI 
No GDI is present. 
DDERR_NOHWND 
Clipper notification requires a window handle, or no window handle has been previously set as the cooperative level window handle. 
DDERR_NOMIPMAPHW 
No mipmap-capable texture mapping hardware is present or available. 
DDERR_NOMIRRORHW 
No mirroring hardware is present or available. 
DDERR_NOMONITORINFORMATION 
New for DirectX 7.0. Testing cannot proceed because the monitor has no associated EDID data.
DDERR_NONONLOCALVIDMEM 
An attempt was made to allocate nonlocal video memory from a device that does not support nonlocal video memory. 
DDERR_NOOPTIMIZEHW 
The device does not support optimized surfaces. 
DDERR_NOOVERLAYDEST
The IDirectDrawSurface7::GetOverlayPosition method is called on an overlay that the IDirectDrawSurface7::UpdateOverlay method has not been called on to establish as a destination.
DDERR_NOOVERLAYHW
No overlay hardware is present or available. 
DDERR_NOPALETTEATTACHED
No palette object is attached to this surface. 
DDERR_NOPALETTEHW 
There is no hardware support for 16- or 256-color palettes.
DDERR_NORASTEROPHW 
No appropriate raster-operation hardware is present or available.
DDERR_NOROTATIONHW 
No rotation hardware is present or available. 
DDERR_NOSTEREOHARDWARE 
There is no stereo hardware present or available. 
DDERR_NOSTRETCHHW 
There is no hardware support for stretching. 
DDERR_NOSURFACELEFT 
There is no hardware present that supports stereo surfaces.
DDERR_NOT4BITCOLOR 
The DirectDrawSurface object is not using a 4-bit color palette, and the requested operation requires a 4-bit color palette. 
DDERR_NOT4BITCOLORINDEX
The DirectDrawSurface object is not using a 4-bit color index palette, and the requested operation requires a 4-bit color index palette. 
DDERR_NOT8BITCOLOR 
The DirectDrawSurface object is not using an 8-bit color palette, and the requested operation requires an 8-bit color palette.
DDERR_NOTAOVERLAYSURFACE 
An overlay component is called for a nonoverlay surface. 
DDERR_NOTEXTUREHW 
The operation cannot be carried out because no texture-mapping hardware is present or available. 
DDERR_NOTFLIPPABLE 
An attempt was made to flip a surface that cannot be flipped. 
DDERR_NOTFOUND
The requested item was not found. 
DDERR_NOTINITIALIZED 
An attempt was made to call an interface method of a DirectDraw object created by CoCreateInstance before the object was initialized. 
DDERR_NOTLOADED 
The surface is an optimized surface, but it has not yet been allocated any memory.
DDERR_NOTLOCKED
An attempt was made to unlock a surface that was not locked. 
DDERR_NOTPAGELOCKED 
An attempt was made to page-unlock a surface with no outstanding page locks. 
DDERR_NOTPALETTIZED 
The surface being used is not a palette-based surface. 
DDERR_NOVSYNCHW
There is no hardware support for vertical blanksynchronized operations.
DDERR_NOZBUFFERHW 
The operation to create a z-buffer in display memory or to perform a blit, using a z-buffer cannot be carried out because there is no hardware support for z-buffers. 
DDERR_NOZOVERLAYHW 
The overlay surfaces cannot be z-layered, based on the z-order because the hardware does not support z-ordering of overlays. 
DDERR_OUTOFCAPS 
The hardware needed for the requested operation has already been allocated.
DDERR_OUTOFMEMORY 
DirectDraw does not have enough memory to perform the operation. 
DDERR_OUTOFVIDEOMEMORY 
DirectDraw does not have enough display memory to perform the operation. 
DDERR_OVERLAPPINGRECTS 
The source and destination rectangles are on the same surface and overlap each other.
DDERR_OVERLAYCANTCLIP 
The hardware does not support clipped overlays. 
DDERR_OVERLAYCOLORKEYONLYONEACTIVE 
An attempt was made to have more than one color key active on an overlay. 
DDERR_OVERLAYNOTVISIBLE 
The IDirectDrawSurface7::GetOverlayPosition method was called on a hidden overlay. 
DDERR_PALETTEBUSY
Access to this palette is refused because the palette is locked by another thread. 
DDERR_PRIMARYSURFACEALREADYEXISTS 
This process has already created a primary surface.
DDERR_REGIONTOOSMALL 
The region passed to the IDirectDrawClipper::GetClipList method is too small. 
DDERR_SURFACEALREADYATTACHED 
An attempt was made to attach a surface to another surface to which it is already attached. 
DDERR_SURFACEALREADYDEPENDENT
An attempt was made to make a surface a dependency of another surface on which it is already dependent.
DDERR_SURFACEBUSY 
Access to the surface is refused because the surface is locked by another thread. 
DDERR_SURFACEISOBSCURED 
Access to the surface is refused because the surface is obscured. 
DDERR_SURFACELOST 
Access to the surface is refused because the surface memory is gone. Call the IDirectDrawSurface7::Restore method on this surface to restore the memory associated with it.
DDERR_SURFACENOTATTACHED
The requested surface is not attached.
DDERR_TESTFINISHED 
New for DirectX 7.0. When returned by the IDirectDraw7::StartModeTest method, this value means that no test could be initiated because all the resolutions chosen for testing already have refresh rate information in the registry. When returned by IDirectDraw7::EvaluateMode, the value means that DirectDraw has completed a refresh rate test.
DDERR_TOOBIGHEIGHT 
The height requested by DirectDraw is too large. 
DDERR_TOOBIGSIZE 
The size requested by DirectDraw is too large. However, the individual height and width are valid sizes. 
DDERR_TOOBIGWIDTH
The width requested by DirectDraw is too large.
DDERR_UNSUPPORTED
The operation is not supported.
DDERR_UNSUPPORTEDFORMAT
The pixel format requested is not supported by DirectDraw.
DDERR_UNSUPPORTEDMASK
The bitmask in the pixel format requested is not supported by DirectDraw.
DDERR_UNSUPPORTEDMODE
The display is currently in an unsupported mode.
DDERR_VERTICALBLANKINPROGRESS
A vertical blank is in progress.
DDERR_VIDEONOTACTIVE
The video port is not active.
DDERR_WASSTILLDRAWING
The previous blit operation that is transferring information to or from this surface is incomplete.
DDERR_WRONGMODE
This surface cannot be restored because it was created in a different mode.
DDERR_XALIGN
The provided rectangle was not horizontally aligned on a required boundary.}
  else
    Result:=Format('DirectDraw error %.8x',[Error]);
  end;
end;

procedure CreateInterfaceFailed(Result: DWord);
begin
  case CoCreateInstance(CLSID_DirectSound,nil,CLSCTX_INPROC_SERVER,IID_IDirectSound,Result) of
    S_OK : ;
    REGDB_E_CLASSNOTREG : raise Exception.Create('Class is not registered in the registration database');
    E_OUTOFMEMORY : raise Exception.Create('Out of memory');
    E_INVALIDARG : raise Exception.Create('One or more arguments are invalid');
    E_UNEXPECTED : raise Exception.Create('Unexpected error occurred');
    CLASS_E_NOAGGREGATION : raise Exception.Create('Class cannot be created as part of an aggregate');
  else
    raise Exception.Create('Error creating interface');
  end;
end;

function GetDirectDrawInterface: IDirectDraw;
begin
  CreateInterfaceFailed(CoCreateInstance(CLSID_DirectDraw,nil,CLSCTX_INPROC_SERVER,IID_IDirectDraw,Result));
  DDFailed(Result.Initialize(nil));
end;

function GetDirectDraw4Interface: IDirectDraw4;
begin
  CreateInterfaceFailed(CoCreateInstance(CLSID_DirectDraw,nil,CLSCTX_INPROC_SERVER,IID_IDirectDraw4,Result));
  DDFailed(Result.Initialize(nil));
end;

function GetDirectDraw7Interface: IDirectDraw7;
begin
  CreateInterfaceFailed(CoCreateInstance(CLSID_DirectDraw7,nil,CLSCTX_INPROC_SERVER,IID_IDirectDraw7,Result));
  DDFailed(Result.Initialize(nil));
end;

procedure DDSurfaceFillColor(const Surface: IDirectDrawSurface7; Color: DWord);
var
  PaintFX : DDBLTFX;
begin
  PaintFX.dwSize:=SizeOf(PaintFX);
  PaintFX.dwFillColor:=Color;
  DDFailed(Surface.Blt(TRect(nil^),nil,TRect(nil^),DDBLT_COLORFILL or DDBLT_WAIT,PaintFX));
end;

procedure DrawSurfaceToLinearBitmapGray(Surface: IDirectDrawSurface7; Image: TLinearBitmap);
var
  X, Y : Integer;
  SrcPix, DestPix : PByte;
  SurfaceDesc : TDDSurfaceDesc2;  // Surface description
  PixSize : Integer;
begin
  SurfaceDesc.dwSize:=SizeOf(SurfaceDesc);
  if Surface.Lock(nil,SurfaceDesc,DDLOCK_WAIT or DDLOCK_READONLY,0)=DD_OK then
  try
    Image.New(SurfaceDesc.dwWidth,SurfaceDesc.dwHeight,pf8bit);
    PixSize:=SurfaceDesc.ddpfPixelFormat.dwRGBBitCount div 8;
    for Y:=0 to SurfaceDesc.dwHeight-1 do
    begin
      SrcPix:=Pointer(Integer(SurfaceDesc.lpSurface)+Y*SurfaceDesc.lPitch);
      DestPix:=Image.ScanLine[Y];
      for X:=1 to SurfaceDesc.dwWidth do
      begin
        DestPix^:=SrcPix^;
        Inc(SrcPix,PixSize);
        Inc(DestPix);
      end;
    end
  finally
    Surface.UnLock(nil);
  end;
end;

procedure DrawSurfaceToLinearBitmapRGB(Surface: IDirectDrawSurface7; Image: TLinearBitmap);
var
  X, Y : Integer;
  SrcPix  : PByte;
  DestPix : PRGBRec;
  SurfaceDesc : TDDSurfaceDesc2;  // Surface description
  PixSize : Integer;
begin
  SurfaceDesc.dwSize:=SizeOf(SurfaceDesc);
  if Surface.Lock(nil,SurfaceDesc,DDLOCK_WAIT or DDLOCK_READONLY,0)=DD_OK then
  try
    Image.New(SurfaceDesc.dwWidth,SurfaceDesc.dwHeight,pf24bit);
    PixSize:=SurfaceDesc.ddpfPixelFormat.dwRGBBitCount div 8;
    if PixSize<3 then raise Exception.Create(rsInvalidPixelFormat);
    for Y:=0 to SurfaceDesc.dwHeight-1 do
    begin
      SrcPix:=Pointer(Integer(SurfaceDesc.lpSurface)+Y*SurfaceDesc.lPitch);
      DestPix:=Image.ScanLine[Y];
      for X:=1 to SurfaceDesc.dwWidth do
      begin
        DestPix^:=PRGBRec(SrcPix)^;
        Inc(SrcPix,PixSize);
        Inc(DestPix);
      end;
    end
  finally
    Surface.UnLock(nil);
  end;
end;

{D-}
procedure DDFailed(Result: Integer);
begin
  if Result<>DD_OK then raise EDirectDrawError.Create(DDErrorString(Result));
end;

initialization
  CoInitialize(nil);
finalization
  CoUninitialize();
end.
