////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// FastBitmap.pas - TBitmap descendant with fast access to bitmap memory
// ---------------------------------------------------------------------
// Changed:   2002-09-24
// Maintain:  Solo: http://meesoft.cjb.net  ||  solo@meesoft.cjb.net
//
// Last change:
//
unit FastBitmap;

interface

uses SysUtils, Windows, Monitor, Graphics, MemUtils, MathUtils;

type
  TFastBitmap = class(TBitmap)
    private
      FWidth, FHeight : Integer;
      function GetScanLine32(Y: Integer): PIntegerArray;
    public
      BytesPerLine : Integer;
      Map : PByteArray;
      Map32 : PIntegerArray;
      property Width: Integer read FWidth;
      property Height: Integer read FHeight;
      property ScanLine32[Y: Integer]: PIntegerArray read GetScanLine32;
      procedure UpdateMap;
      procedure Clear;
      // pfCustom means no change
      procedure New(X,Y: Integer; PixFormat: TPixelFormat=pfCustom);

      procedure Smooth(WindowSize: Integer);
    end;

implementation

//==============================================================================================================================
// TFastBitmap
//==============================================================================================================================
procedure TFastBitmap.UpdateMap;
begin
  FWidth:=inherited Width;
  FHeight:=inherited Height;
  case PixelFormat of
    pf8bit  : BytesPerLine:=Ceil4(Width);
    pf24bit : BytesPerLine:=Ceil4(Width*3);
    pf32bit : BytesPerLine:=Ceil4(Width*4);
  else raise Exception.Create('Unsupported pixel format');
  end;
  Map:=ScanLine[Height-1];
  Map32:=Pointer(Map);
end;

procedure TFastBitmap.New(X,Y: Integer; PixFormat: TPixelFormat=pfCustom);
begin
  inherited Width:=X;
  inherited Height:=Y;
  if PixFormat<>pfCustom then PixelFormat:=PixFormat;
  UpdateMap;
end;

procedure TFastBitmap.Clear;
var
  Y : Integer;
begin
  for Y:=0 to Height-1 do ZeroMem(Pointer(Integer(Map)+Y*BytesPerLine)^,Width*4);
  //ZeroMem(Map^,Height*BytesPerLine);
end;

function TFastBitmap.GetScanLine32(Y: Integer): PIntegerArray;
begin
  Result:=PIntegerArray(Integer(Map)+Y*BytesPerLine);
end;

procedure TFastBitmap.Smooth(WindowSize: Integer);
var
  NewMap : PIntegerArray;
  X, Y, DX, DY, DMax, DMin, XA, XB, YA, YB, P  : Integer;
  CenterPix, ScanPix : PByte;
  Count, Sum : Integer;
begin
  Assert(PixelFormat=pf32bit);

  DMin:=-WindowSize div 2;
  DMax:=WindowSize+DMin-1;

  GetMem(NewMap,BytesPerLine*Height);
  for P:=0 to 2 do
  begin
    CenterPix:=Pointer(Integer(NewMap)+P);
    for Y:=0 to Height-1 do
    begin
      YA:=Y+DMin; if YA<0 then YA:=0;
      YB:=Y+DMax; if YB>=Height then YB:=Height-1;
      for X:=0 to Width-1 do
      begin
        Count:=0; Sum:=0;
        XA:=X+DMin; if XA<0 then XA:=0;
        XB:=X+DMax; if XB>=Width then XB:=Width-1;
        for DY:=YA to YB do
        begin
          ScanPix:=@Map^[DY*BytesPerLine+XA*4+P];
          for DX:=XA to XB do
          begin
            Sum:=Sum+ScanPix^;
            Inc(Count);
            Inc(ScanPix,4);
          end;
        end;
        CenterPix^:=Sum div Count;
        Inc(CenterPix,4);
      end;
    end;
  end;
  Move(NewMap^,Map^,BytesPerLine*Height);
  FreeMem(NewMap);
end;

end.

