////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// PlayThread.pas - MPEG2 player thread
// ------------------------------------
// Version:   2005-01-09
// Maintain:  Michael Vinther    |    mv@logicnetdk
//
unit PlayThread;

interface

uses
  Windows, Classes, SysUtils, MPEG2Lib, DirectX, DirectDrawUtils, MMSystem;

type
  TPlayEvent = set of (peFrameReady,peError,peStop,peStatsUpdated,peEndOfStream);
  TPlayThreadEvent = procedure(State: TPlayEvent) of object;

  TMPEG2PlayThread = class(TThread)
    protected
      State : TPlayEvent;
      Decoder : TMPEG2Decoder;
      procedure ShowFrame;
      procedure Execute; override;
    public
      FrameBuffer : IDirectDrawSurface7;
      FrameReady : TPlayThreadEvent;
      DecoderFramesPerSecond : Integer;
      FrameDelay : Integer; // Minimum time between frames (ms)
      DeInterlacingStrength : Integer;
      constructor Create(MPEG2Decoder: TMPEG2Decoder);
      destructor Destroy; override;
    end;

  TRenderFrame = procedure(Frame: PByteArray; FrameBuffer: IDirectDrawSurface7);

var
  RenderFrame : TRenderFrame;

implementation

//==============================================================================================================================
// TMPEG2PlayThread
//==============================================================================================================================
constructor TMPEG2PlayThread.Create(MPEG2Decoder: TMPEG2Decoder);
begin
  inherited Create(True);
  Decoder:=MPEG2Decoder;
end;

destructor TMPEG2PlayThread.Destroy;
begin
  FrameReady:=nil;
  inherited;
end;

procedure TMPEG2PlayThread.ShowFrame;
begin
  if Assigned(FrameReady) then FrameReady(State);
end;

procedure TMPEG2PlayThread.Execute;
var
  Frame : PByteArray;
  Time : DWord;
  LastUpdateTime, FrameCount : DWord;
begin
  try
    FrameCount:=0;
    LastUpdateTime:=timeGetTime;
    while not Terminated do
    begin
      Frame:=Decoder.DecodeFrame(DeInterlacingStrength);
      if Assigned(Frame) then
      begin
        RenderFrame(Frame,FrameBuffer);
        State:=[peFrameReady];

        Inc(FrameCount);
        Time:=timeGetTime;
        if Time>=LastUpdateTime+1000 then
        begin
          DecoderFramesPerSecond:=Round(FrameCount*1000/(Time-LastUpdateTime));
          FrameCount:=0;
          LastUpdateTime:=Time;
          State:=State+[peStatsUpdated];
        end;
        if Assigned(FrameReady) then Synchronize(ShowFrame);
      end
      else // No more frames
      begin
        State:=[peStop,peEndOfStream];
        if Assigned(FrameReady) then Synchronize(ShowFrame);
        Break;
      end;
    end;
  except
    State:=[peStop,peError];
    if Assigned(FrameReady) then Synchronize(ShowFrame);
  end;
end;

end.

