unit ViewMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, MPEG2Lib, MemUtils, Buttons, DirectX, LinarBitmap,
  ComCtrls, PlayThread, MMSystem, DirectDrawPanel, DirectDrawUtils, Monitor,
  FileUtils, About, Menus, MathUtils, ShellAPI, Math;

const
  TrackBlockSize = 10*1024;

type
  TMainForm = class(TForm)
    ControlPanel: TPanel;
    TrackBar: TTrackBar;
    InfoLabel: TLabel;
    MainMenu: TMainMenu;
    FileMenu: TMenuItem;
    HelpMenu: TMenuItem;
    AboutItem: TMenuItem;
    ExitItem: TMenuItem;
    OpenItem: TMenuItem;
    N1: TMenuItem;
    VideoMenu: TMenuItem;
    CopyFrameItem: TMenuItem;
    ButtonPlay: TButton;
    ButtonPlayFrame: TButton;
    ButtonStop: TButton;
    OpenDiskItem: TMenuItem;
    OpenDialog: TOpenDialog;
    StopItem: TMenuItem;
    LimitFrameRateItem: TMenuItem;
    N2: TMenuItem;
    VideoInformationItem: TMenuItem;
    ButtonSeekStart: TButton;
    ColorModeMenu: TMenuItem;
    ColorModeGrayscaleItem: TMenuItem;
    ColorModeRGBItem: TMenuItem;
    ExtractTrackDataItem: TMenuItem;
    N3: TMenuItem;
    SetTrackBarRangeItem: TMenuItem;
    DirectDrawPanel: TDirectDrawPanel;
    N4: TMenuItem;
    DeinterlacingItem: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure ButtonPlayFrameClick(Sender: TObject);
    procedure TrackBarChange(Sender: TObject);
    procedure ButtonPlayClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ButtonStopClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure AboutItemClick(Sender: TObject);
    procedure ExitItemClick(Sender: TObject);
    procedure OpenItemClick(Sender: TObject);
    procedure CopyFrameItemClick(Sender: TObject);
    procedure LimitFrameRateItemClick(Sender: TObject);
    procedure VideoInformationItemClick(Sender: TObject);
    procedure ButtonSeekStartClick(Sender: TObject);
    procedure ColorModeItemClick(Sender: TObject);
    procedure ExtractTrackDataItemClick(Sender: TObject);
    procedure SetTrackBarRangeItemClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure ButtonKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure TrackBarKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure DeinterlacingItemClick(Sender: TObject);
  private
    { Private declarations }
    PlayThread : TMPEG2PlayThread;
    MPEG2Decoder : TMPEG2Decoder;
    NextFrameTime, MinimumFrameDelay : DWord;
    DeInterlacingStrength : Integer;
    procedure FrameReady(State: TPlayEvent);
    procedure Play;
    procedure WMDropFiles(var Msg: TMessage); message WM_DROPFILES;
  public
    { Public declarations }
    FileName : string;
    Playing : Boolean;
    procedure UpdateControlStates;
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

procedure TMainForm.FormCreate(Sender: TObject);
begin
  Caption:=Application.Title;
  InfoLabel.Caption:='';
  ControlPanel.Enabled:=False;
  LongTimeFormat:='hh:nn:ss';
  //MPEG2Decoder:=TMPEG2Decoder.Create('C:\PROGRM\C\MPEG2Decode\MPEG2Lib\Release');
  MPEG2Decoder:=TMPEG2Decoder.Create;

  if not (GetDeviceCaps(Canvas.Handle,BITSPIXEL) in [24,32]) then
    raise Exception.Create('Display mode not supported: Must be 24 or 32 bit');
  ControlPanel.Enabled:=True;
  DeInterlacingStrength:=240;
  ColorModeItemClick(ColorModeRGBItem);
  OpenDiskItem.Visible:=Win32Platform>=VER_PLATFORM_WIN32_NT;
  UpdateControlStates;
  DragAcceptFiles(Handle,True);
end;

procedure TMainForm.FormShow(Sender: TObject);
begin
  DirectDrawPanel.StretchBackBuffer(16,16);
  DDSurfaceFillColor(DirectDrawPanel.DDSBackBuffer,0);
  if ParamCount>0 then
  begin
    FileName:=GetParameterFileName;
    OpenItemClick(Self);
  end;
end;

procedure TMainForm.ButtonPlayFrameClick(Sender: TObject);
var
  Frame : PByteArray;
begin
  if Playing then ButtonStopClick(Self);
  if Sender=ButtonPlayFrame then InfoLabel.Caption:='';
  Frame:=MPEG2Decoder.DecodeFrame(DeInterlacingStrength);
  if Assigned(Frame) then
  begin
    RenderFrame(Frame,DirectDrawPanel.DDSBackBuffer);
    DirectDrawPanel.Invalidate;
  end;
end;

procedure TMainForm.Play;
begin
  PlayThread.Free;
  NextFrameTime:=0;
  PlayThread:=TMPEG2PlayThread.Create(MPEG2Decoder);
  PlayThread.DeInterlacingStrength:=DeInterlacingStrength;
  PlayThread.FrameBuffer:=DirectDrawPanel.DDSBackBuffer;
  PlayThread.FrameReady:=FrameReady;
  Playing:=True;
  PlayThread.Resume;
end;

procedure TMainForm.FrameReady(State: TPlayEvent);
var
  Time : DWord;
begin
  if peFrameReady in State then
  begin
    if MinimumFrameDelay>0 then
    begin
      repeat
        Time:=timeGetTime;
      until Time>=NextFrameTime;
      Inc(NextFrameTime,MinimumFrameDelay);
      if NextFrameTime<Time then NextFrameTime:=Time;
    end;
    DirectDrawPanel.Invalidate;
  end;
  if ([peStatsUpdated,peEndOfStream]*State<>[]) and Assigned(PlayThread) then
  begin
    InfoLabel.Caption:=TimeToStr(MPEG2Decoder.Timestamp/(90000*24*3600.0))+' ('+IntToStr(PlayThread.DecoderFramesPerSecond)+' fps)';
    TrackBar.OnChange:=nil;
    TrackBar.Position:=MPEG2Decoder.Position div TrackBlockSize;
    TrackBar.OnChange:=TrackBarChange;
  end;
  if peStop in State then
  begin
    Playing:=False;
    UpdateControlStates;
  end;
  if peError in State then
  begin
    InfoLabel.Caption:='Error';
  end;
end;

procedure TMainForm.TrackBarChange(Sender: TObject);
begin
  if Playing then FreeAndNil(PlayThread);
  MPEG2Decoder.Position:=Int64(TrackBar.Position)*TrackBlockSize;
  InfoLabel.Caption:=TimeToStr(MPEG2Decoder.Timestamp/(90000*24*3600.0));
  if Playing then Play
  else
  begin
    ButtonPlayFrameClick(Self);
    Update;
  end;
end;

procedure TMainForm.UpdateControlStates;
var
  FileOpen : Boolean;
begin
  FileOpen:=Assigned(MPEG2Decoder) and (MPEG2Decoder.StreamSize>0);
  ButtonPlay.Enabled:=FileOpen and not Playing;
  ButtonPlayFrame.Enabled:=FileOpen and not Playing;
  ButtonSeekStart.Enabled:=FileOpen;
  TrackBar.Enabled:=FileOpen;
  VideoInformationItem.Enabled:=FileOpen;
  CopyFrameItem.Enabled:=FileOpen;
  ButtonStop.Enabled:=Playing;
  StopItem.Enabled:=Playing;
  ExtractTrackDataItem.Enabled:=FileOpen;
end;

procedure TMainForm.ButtonPlayClick(Sender: TObject);
begin
  Play;
  UpdateControlStates;
end;

procedure TMainForm.ButtonStopClick(Sender: TObject);
begin
  FreeAndNil(PlayThread);
  Playing:=False;
  UpdateControlStates;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  FreeAndNil(PlayThread);
end;

procedure TMainForm.AboutItemClick(Sender: TObject);
begin
  InfoLabelText:='Based on the DVD2AVI project and MSSG MPEG2Decode.';
  ShowAboutBox;
end;

procedure TMainForm.ExitItemClick(Sender: TObject);
begin
  Close;
end;

var
  VideoFile : File;
  LastPos : Integer;

function StreamGetCallback(Offset: Int64; Buffer: Pointer; Count: DWord): DWord; stdcall;
begin
  Seek(VideoFile,Offset);
  BlockRead(VideoFile,Buffer^,Count,Result);
  LastPos:=Offset+Result;
end;

procedure TMainForm.OpenItemClick(Sender: TObject);
var
  DiskNumber : Integer;
  Str : string;
begin
  if Playing then ButtonStopClick(Self);
  InfoLabel.Caption:='';
  try
    if Sender=Self then MPEG2Decoder.OpenFile(FileName)
    else if Sender=OpenDiskItem then
    begin
      Str:='1';
      if not InputQuery('Open disk','Enter disk number (first HD is 0):',Str) then Exit;
      DiskNumber:=StrToInt(Str);
      FileName:='[Disk '+IntToStr(DiskNumber)+']';
      MPEG2Decoder.OpenDisk(DiskNumber,1024*1024);
    end
    else // Standard open dialog
    begin
      if (FileName<>'') and (FileName[1]<>'[') then OpenDialog.FileName:=FileName;
      if not OpenDialog.Execute then Exit;
      FileName:=OpenDialog.FileName;

      MPEG2Decoder.OpenFile(FileName);
      {AssignFile(VideoFile,FileName);
      Reset(VideoFile,1); LastPos:=0;
      MPEG2Decoder.OpenStream(StreamGetCallback,True,FileSize(VideoFile));{}
    end;
    if MPEG2Decoder.AspectRatio<>0 then ClientWidth:=Round(MPEG2Decoder.FrameHeight*MPEG2Decoder.AspectRatio)
    else ClientWidth:=MPEG2Decoder.FrameWidth;
    if MPEG2Decoder.FrameRate>0 then MinimumFrameDelay:=ForceInRange(Round(1000/MPEG2Decoder.FrameRate),0,1000);
    ClientHeight:=ControlPanel.Height+MPEG2Decoder.FrameHeight;
    Left:=(Screen.Width-Width) div 2;
    Top:=(Screen.Height-Height) div 2;
    DirectDrawPanel.StretchBackBuffer(MPEG2Decoder.FrameWidth,MPEG2Decoder.FrameHeight);
    TrackBar.Position:=0;
    TrackBar.Max:=MPEG2Decoder.StreamSize div TrackBlockSize;
    Caption:=Application.Title+' - '+ExtractFileName(FileName);

    ButtonPlayFrameClick(Self);
  finally
    UpdateControlStates;
    if ButtonPlay.Enabled then ButtonPlay.SetFocus;
  end;
end;

procedure TMainForm.CopyFrameItemClick(Sender: TObject);
var
  Image : TLinearBitmap;
begin
  Image:=TLinearBitmap.Create;
  try
    if ColorModeRGBItem.Checked then DrawSurfaceToLinearBitmapRGB(DirectDrawPanel.DDSBackBuffer,Image)
    else
    begin
      DrawSurfaceToLinearBitmapGray(DirectDrawPanel.DDSBackBuffer,Image);
      Image.Palette^:=GrayPal;
    end;
    Image.CopyToClipboard;
  finally
    Image.Free;
  end;
end;

procedure TMainForm.LimitFrameRateItemClick(Sender: TObject);
var
  Str : string;
begin
  if MinimumFrameDelay=0 then Str:='0'
  else Str:=IntToStr(1000 div MinimumFrameDelay);
  if InputQuery(StripHotkey(TMenuItem(Sender).Caption),'Enter desired frame rate (0 for max):',Str) then
  begin
    MinimumFrameDelay:=StrToInt(Str);
    if MinimumFrameDelay<>0 then MinimumFrameDelay:=ForceInRange(1000 div MinimumFrameDelay,0,1000);
  end;
end;

procedure TMainForm.VideoInformationItemClick(Sender: TObject);
begin
  MessageDlg('Resolution: '#9+IntToStr(MPEG2Decoder.FrameWidth)+' x '+IntToStr(MPEG2Decoder.FrameHeight)+' pixels'#13+
             'Aspect ratio: '#9+FloatToStrF(MPEG2Decoder.AspectRatio,ffGeneral,3,0)+#13+
             'Frame rate: '#9+FloatToStrF(MPEG2Decoder.FrameRate,ffGeneral,6,0)+' fps',
             mtInformation,[mbOk],0);
end;

procedure TMainForm.ButtonSeekStartClick(Sender: TObject);
begin
  TrackBar.Position:=0;
end;

procedure TMainForm.ColorModeItemClick(Sender: TObject);
begin
  FreeAndNil(PlayThread);
  if Sender=ColorModeGrayscaleItem then
  begin
    MPEG2Decoder.PixelFormat:=vpfGray8;
    RenderFrame:=RenderFrameGray
  end
  else if Sender=ColorModeRGBItem then
  begin
    MPEG2Decoder.PixelFormat:=vpfRGB24;
    RenderFrame:=RenderFrame24Flip;
  end
  else raise Exception.Create('Unsupported pixel format');
  if Playing then Play;
  TMenuItem(Sender).Checked:=True;
end;

procedure TMainForm.ExtractTrackDataItemClick(Sender: TObject);
var
  Str : string;
  Size : Int64;
begin
  if Playing then ButtonStopClick(Self);
  Str:='10';
  if not InputQuery(StripHotkey(ExtractTrackDataItem.Caption),'Size of stream to extract (Mb):',Str) then Exit;
  Size:=Int64(StrToInt(Str))*(1024*1024);
  if Size<1 then Exit;

  Str:='c:\temp\stream.mpeg2';
  if not InputQuery(StripHotkey(ExtractTrackDataItem.Caption),'Filename:',Str) then Exit;

  Screen.Cursor:=crHourGlass;
  try
    Size:=MPEG2Decoder.WriteToFile(PChar(Str),Size);
  finally
    Screen.Cursor:=crDefault;
  end;
  if Size<=0 then MessageDlg('Unable to create file: '+Str,mtError,[mbOk],0);
end;

procedure TMainForm.WMDropFiles(var Msg: TMessage);
begin
  try
    if DragQueryFile(Msg.wParam,$FFFFFFFF,nil,0)>0 then
    begin
      SetLength(FileName,DragQueryFile(Msg.wParam,0,nil,MAX_PATH));
      SetLength(FileName,DragQueryFile(Msg.wParam,0,PChar(FileName),MAX_PATH));
      OpenItemClick(Self);
    end;
  finally
    DragFinish(Msg.wParam);
  end;
end;

procedure TMainForm.SetTrackBarRangeItemClick(Sender: TObject);
var
  Str : string;
begin
  Str:=IntToStr(TrackBar.Max*(TrackBlockSize div 1024) div 1024);
  if not InputQuery(StripHotkey(SetTrackBarRangeItem.Caption),'Range (Mb):',Str) then Exit;
  TrackBar.Max:=Min(MPEG2Decoder.StreamSize div TrackBlockSize,StrToInt(Str)*1024 div (TrackBlockSize div 1024));
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  MPEG2Decoder.Free;
end;

procedure TMainForm.ButtonKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if Key=VK_CONTROL then TrackBar.SetFocus;
end;

procedure TMainForm.TrackBarKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if (Key=VK_SPACE) and ButtonPlay.Enabled then ButtonPlayClick(Self);
end;

procedure TMainForm.DeinterlacingItemClick(Sender: TObject);
var
  Str : string;
begin
  Str:=IntToStr(DeInterlacingStrength);
  if InputQuery(StripHotkey(TMenuItem(Sender).Caption),'Filter strength:',Str) then
  begin
    DeInterlacingStrength:=StrToInt(Str);
    if Assigned(PlayThread) then PlayThread.DeInterlacingStrength:=DeInterlacingStrength;
  end;
end;

end.

