unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, ExtCtrls, StdCtrls, MPEG2Lib, Math;

type
  TForm1 = class(TForm)
    OpenDialog1: TOpenDialog;
    OpenButton: TButton;
    Image1: TImage;
    PlayButton: TButton;
    Timer1: TTimer;
    StopButton: TButton;
    procedure OpenButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure PlayButtonClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    Stream : TMemoryStream;
    MPEG2Decoder : TMPEG2Decoder;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

procedure TForm1.FormCreate(Sender: TObject);
begin
  MPEG2Decoder:=TMPEG2Decoder.Create;
end;

procedure TForm1.FormDestroy(Sender: TObject);
begin
  MPEG2Decoder.Free;
  Stream.Free;
end;

function StreamGetCallback(Offset: Int64; Buffer: Pointer; Count: DWord; Owner: TObject): DWord; stdcall;
begin
  Result:=Min(Count,Form1.Stream.Size-Offset);
  Move(PByteArray(Form1.Stream.Memory)^[Offset],Buffer^,Result);
end;

procedure TForm1.OpenButtonClick(Sender: TObject);

begin
  if not OpenDialog1.Execute then Exit;
  Stream.Free;

  Stream:=TMemoryStream.Create;
  Stream.LoadFromFile(OpenDialog1.FileName);

  MPEG2Decoder.OpenStream(StreamGetCallback,True,Stream.Size,Self);

  MPEG2Decoder.PixelFormat:=vpfRGB24;

  Image1.Height:=MPEG2Decoder.FrameHeight;
  Image1.Width:=MPEG2Decoder.FrameHeight*4 div 3;

  ClientWidth:=Image1.Width+2*Image1.Left;
  ClientHeight:=Image1.Height+Image1.Top+Image1.Left;

  Image1.Picture.Bitmap.PixelFormat:=pf24bit;
  Image1.Picture.Bitmap.Width:=MPEG2Decoder.FrameWidth;
  Image1.Picture.Bitmap.Height:=MPEG2Decoder.FrameHeight;

  Timer1.Interval:=Round(1000/MPEG2Decoder.FrameRate);

  PlayButton.Enabled:=True;
  Timer1Timer(nil);
end;

procedure TForm1.PlayButtonClick(Sender: TObject);
begin
  MPEG2Decoder.Position:=0;
  Timer1.Enabled:=True;
  PlayButton.Enabled:=False;
  StopButton.Enabled:=True;
end;

procedure TForm1.StopButtonClick(Sender: TObject);
begin
  Timer1.Enabled:=False;
  PlayButton.Enabled:=True;
  StopButton.Enabled:=False;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
  Frame : PByteArray;
  Y : Integer;
begin
  Frame:=MPEG2Decoder.DecodeFrame;
  if Assigned(Frame) then
  begin
    for Y:=0 to MPEG2Decoder.FrameHeight-1 do
      Move(Frame^[Y*MPEG2Decoder.FrameWidth*3],
           Image1.Picture.Bitmap.ScanLine[MPEG2Decoder.FrameHeight-1-Y]^,
           MPEG2Decoder.FrameWidth*3);
    Image1.Repaint;
  end
  else StopButtonClick(nil);
end;

end.

